/*
 * Copyright 2021 Michael Ferguson
 * Copyright 2015 Fetch Robotics Inc
 * Author: Michael Ferguson
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <graceful_controller/graceful_controller.hpp>

#include <angles/angles.h>

#include <algorithm>
#include <list>
#include <vector>
#include <cmath>

namespace graceful_controller
{

GracefulController::GracefulController(double k1, double k2,
                                       double min_velocity, double max_velocity,
                                       double max_accel,
                                       double max_angular_velocity,
                                       double beta, double lambda)
{
  k1_ = k1;
  k2_ = k2;
  min_velocity_ = min_velocity;
  max_velocity_ = max_velocity;
  max_accel_ = max_accel;
  max_angular_velocity_ = max_angular_velocity;
  beta_ = beta;
  lambda_ = lambda;
}

// x, y, theta are relative to base location and orientation
bool GracefulController::approach(const double x, const double y, const double theta,
                                  double& vel_x, double& vel_th)
{
  // Distance to goal
  double r = std::sqrt(x * x + y * y);

  // Orientation base frame relative to r_
  double delta = std::atan2(-y, x);

  // Determine orientation of goal frame relative to r_
  double theta2 = angles::normalize_angle(theta + delta);

  // Compute the virtual control
  double a = std::atan(-k1_ * theta2);
  // Compute curvature (k)
  double k = -1.0/r * (k2_ * (delta - a) + (1 + (k1_/(1+((k1_*theta2)*(k1_*theta2)))))*sin(delta));

  // Compute max_velocity based on curvature
  double v = max_velocity_ / (1 + beta_ * std::pow(fabs(k), lambda_));
  // Limit velocity based on approaching target
  double approach_limit = 2 * max_accel_ * r;
  v = std::min(v, approach_limit);
  v = std::min(std::max(v, min_velocity_), max_velocity_);

  // Compute angular velocity
  double w = k * v;
  // Bound angular velocity
  double bounded_w = std::min(max_angular_velocity_, std::max(-max_angular_velocity_, w));
  // Make sure that if we reduce w, we reduce v so that kurvature is still followed
  if (w != 0.0)
  {
    v *= (bounded_w/w);
  }

  // Send command to base
  vel_x = v;
  vel_th = bounded_w;
  return true;
}

void GracefulController::setVelocityLimits(
  const double min_velocity,
  const double max_velocity,
  const double max_angular_velocity)
{
  min_velocity_ = min_velocity;
  max_velocity_ = max_velocity;
  max_angular_velocity_ = max_angular_velocity;
}

}  // namespace graceful_controller
